/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2015 KYOCERA Corporation
 * (C) 2017 KYOCERA Corporation
 *
 * Copyright (c) 2010-2011, Code Aurora Forum. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 and
 * only version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 *
*/
#include <linux/module.h>
#include <linux/kernel.h>

#include "disp_ext.h"
static int panel_id = -1;

static int detect_panel_id(struct platform_device *pdev)
{
	int id = 0;

	if (panel_id != -1) {
		pr_notice("%s: panel id Checked\n", __func__);
		return panel_id;
	}

	id = 1;

	panel_id = id;
	pr_info("%s: panel id = %d\n", __func__, panel_id);
	return panel_id;
}

struct device_node *disp_ext_find_panel_of_node(struct platform_device *pdev)
{
	struct device_node *dsi_pan_node = NULL, *mdss_node = NULL;

	mdss_node = of_parse_phandle(pdev->dev.of_node,
			"qcom,mdss-mdp", 0);
	if (!mdss_node) {
		pr_err("%s: %d: mdss_node null\n", __func__, __LINE__);
		return NULL;
	}

	switch (detect_panel_id(pdev)) {
	case 0:
		dsi_pan_node = of_find_node_by_name(mdss_node,
				"qcom,mdss_dsi_kc_innolux_qvga_cmd");
		break;
	case 1:
	default:
		dsi_pan_node = of_find_node_by_name(mdss_node,
				"qcom,mdss_dsi_kc_byd_qvga_cmd");
		break;
	};

	return dsi_pan_node;
}

int disp_ext_get_panel_id(void)
{
	return panel_id;
}
